
// fx - An compositing library representation.

// Copyright (C) 2005-2006 VM Inc.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

// simple example to exercise network evaluation.

#include <cstdio>
#include <cstdlib>

#ifdef HAVE_CONFIG_H
#include <openlibraries_global_config.hpp>
#endif

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#include <openeffectslib/fx/openeffectslib_plugin.hpp>
#include <openeffectslib/fx/connect.hpp>

#include <openimagelib/il/openimagelib_plugin.hpp>

#include <GL/glew.h>

#if defined ( WIN32 ) || defined ( HAVE_GL_GLUT_H )
#include <GL/glut.h>
#else
#include <GLUT/glut.h>
#endif

int WIN_WIDTH  = 720;
int WIN_HEIGHT = 576;

namespace il  = olib::openimagelib::il;
namespace oel = olib::openeffectslib;
namespace opl = olib::openpluginlib;
namespace fs  = boost::filesystem;

namespace
{
#ifdef WIN32
	const opl::string plugin_path( "./plugins" );
#else
	const opl::string oil_plugin_path( OPENIMAGELIB_PLUGINS );
	const opl::string oel_plugin_path( OPENEFFECTSLIB_PLUGINS );
	const opl::string oml_plugin_path( OPENMEDIALIB_PLUGINS );
#endif

	struct source_query_traits
	{
		opl::wstring libname( ) const
		{ return L"openeffectslib"; }
		
		opl::wstring to_match( ) const
		{ return L"source/media"; }

		opl::wstring type( ) const
		{ return L""; }
		
		int merit( ) const
		{ return 0; }
	};

	struct sepia_effect_query_traits
	{
		opl::wstring libname( ) const
		{ return L"openeffectslib"; }
		
		opl::wstring to_match( ) const
		{ return L"sepia/effect"; }

		opl::wstring type( ) const
		{ return L""; }
		
		int merit( ) const
		{ return 0; }
	};

	oel::fx::n_operator_ptr op_source, op_effect;
}

static void calculate_dimensions( int phy_w, int phy_h, int& req_w, int& req_h )
{
	// This is the absolute WxH we have
	int abs_w = WIN_WIDTH;
	int abs_h = WIN_HEIGHT;
	
	// This is the physcial WxH of the image
	double ar = ( double ) phy_w / phy_h;

	// Calculate the required display WxH
	req_h = abs_h;
	req_w = int( req_h * ar );
	
	// Scale up or down to fit
	if ( req_w > abs_w )
	{
		req_w = abs_w;
		req_h = int( req_w / ar );
	}
}

static bool create_network( const boost::filesystem::path& path )
{
	typedef opl::discovery<source_query_traits>			oel_source_discovery;
	typedef opl::discovery<sepia_effect_query_traits>	oel_sepia_effect_discovery;
	
	oel_source_discovery source_plugins( ( source_query_traits( ) ) );
	if( source_plugins.empty( ) ) return false;
	
	oel_source_discovery::const_iterator i = source_plugins.begin( );
	boost::shared_ptr<oel::fx::openeffectslib_plugin> plug = boost::shared_dynamic_cast<oel::fx::openeffectslib_plugin>( i->create_plugin( "" ) );
	if( !plug ) return 0;

	op_source = plug->create( L"media source" );
	if( !op_source ) return false;
		
	opl::v_url* url = op_source->pointer<opl::v_url>( L"url" );
	url->get( ) = path;

	if( !op_source->init( ) ) return false;
	
	oel_sepia_effect_discovery effect_plugins( ( sepia_effect_query_traits( ) ) );
	if( effect_plugins.empty( ) ) return false;

	i = effect_plugins.begin( );
	plug = boost::shared_dynamic_cast<oel::fx::openeffectslib_plugin>( i->create_plugin( "" ) );
	if( !plug ) return 0;

	op_effect = plug->create( L"" );
	if( !op_effect ) return false;
	
	oel::fx::connect( op_source, L"OutColor0", op_effect, L"inColor0" );
	
	return true;
}

static void display( void )
{
	glDisable( GL_DEPTH_TEST );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glEnable( GL_DEPTH_TEST );

	op_effect->update_image( );
	
	op_effect->bind_as_GL_texture( );
	
	glViewport( 0, 0, WIN_WIDTH, WIN_HEIGHT );

	glMatrixMode( GL_PROJECTION );
	glPushMatrix( );
	glLoadIdentity( );
	gluOrtho2D( 0.0f, WIN_WIDTH, 0.0f, WIN_HEIGHT );
			
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix( );
	glLoadIdentity( );
	
	int phy_w   = op_effect->width( );
	int phy_h   = op_effect->height( );
	float tex_w = op_effect->max_s( );
	float tex_h = op_effect->max_t( );
	
	int req_w, req_h;
	calculate_dimensions( phy_w, phy_h, req_w, req_h );

	float off_x = 0.5 * ( WIN_WIDTH - req_w );
	float off_y = 0.5 * ( WIN_HEIGHT - req_h );
	glTranslatef( off_x, off_y, 0.0f );

	float zoom_w = ( float ) req_w / ( float ) phy_w; 
	float zoom_h = ( float ) req_h / ( float ) phy_h;
	glScalef( zoom_w, zoom_h, 1.0f );

	glBegin( GL_QUADS );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex2f( 0.0f, phy_h );
		
		glTexCoord2f( tex_w, 0.0f );
		glVertex2f( phy_w, phy_h );
		
		glTexCoord2f( tex_w, tex_h );
		glVertex2f( phy_w, 0.0f );
		
		glTexCoord2f( 0.0f, tex_h );
		glVertex2f( 0.0f, 0.0f );
	glEnd( );

	glMatrixMode( GL_PROJECTION );
	glPopMatrix( );

	glMatrixMode( GL_MODELVIEW );
	glPopMatrix( );
	
	op_effect->unbind_as_GL_texture( );
	
#ifndef NDEBUG
	GLenum error;
	while( ( error = glGetError( ) ) != GL_NO_ERROR )
		fprintf( stderr, "GLerror:%s\n", gluErrorString( error ) );
#endif

	oel::fx::v_frame_position *position = op_source->pointer<oel::fx::v_frame_position>( L"imagePosition" );
	position->get( ) = position->value( ) + 1;

	glutSwapBuffers( );
}

static void idle( void )
{
	glutPostRedisplay( );
}

static void reshape( int w, int h )
{
	glViewport( 0, 0, w, h );

	WIN_WIDTH  = w;
	WIN_HEIGHT = h;
}

static void keyboard( unsigned char key, int /*x*/, int /*y*/ )
{
	switch( key )
	{
		case 27:
			op_source = oel::fx::n_operator_ptr( );
			op_effect = oel::fx::n_operator_ptr( );
			exit( 0 );
			break;

		default:
			break;
	}
}

int main( int argc, char* argv[ ] )
{
	glutInit( &argc, argv );
	glutInitWindowSize( WIN_WIDTH, WIN_HEIGHT );
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA );
	glutCreateWindow( "OpenGL Shading Network Evaluation" );
	glutDisplayFunc( display );
	glutIdleFunc( idle );
	glutReshapeFunc( reshape );
	glutKeyboardFunc( keyboard );

	glewInit( );
	
#ifdef WIN32
	opl::init( plugin_path );
#else
	opl::init( oil_plugin_path );
	opl::init( oel_plugin_path );
	opl::init( oml_plugin_path );
#endif
	
	if( argc > 1 )
	{
		if( !create_network( fs::path( argv[ 1 ], fs::native ) ) ) return 0;
		
		glutMainLoop( );
	}
}

